function [xGN, k, rho, X, Alphas] = tiklgn( fun, b, L, x0, lam, niter, mnflag, xbar, curegsvd)
%TIKLGN Gauss-Newton method regularized by Tikhonov in general form
%   [x, k, rho, X, alpha] = tiklgn( fun, b, L, x0, lam, niter, mnflag, xbar)
%   computes the regularized solution to the nonlinear least-square
%   problem min||fun(x)-b||, i.e.,
%   min { ||fun(x)-b||^2 + lam^2||L(x-xbar)||^2 }.
%   where
%      fun      - function and Jacobian: [f,J]=fun(x)
%      b        - data vector
%      L        - regularization matrix
%      x0       - initial point
%      lam      - regularization parameter
%      niter    - max number of iterations
%      mnflag   - 1 minimum norm solution, 0 standard solution
%      xbar     - a priori estimate of solution (only if mnflag=1)
%   and
%      x        - solution
%      k        - number of iterations performed
%      rho      - solution residual
%      X        - solution at each iteration
%      alpha    - step-lenghts for damped Gauss-Newton method

%   F. Pes and G. Rodriguez
%   University of Cagliari, Italy

% Last revised July 24, 2020

if nargin<9, curegsvd = 1; end
if nargin<8, xbar = zeros(size(x0)); end
if nargin<7, mnflag = 1; end
if nargin<6, niter = 40; end

if lam<0, error('Illegal regularization parameter lambda.'), end
if mnflag==0 && ~(isequal(xbar,zeros(size(x0)))), error('xbar only if mnflag=1.'), end

minpowAG = -30;		% alpha > 2^minpowAG for Armijo-Goldstein
factgsvd = 1e-6;

n = size(x0,1);
m = size(b,1);
r = min(m,n);
p = size(L,1);
d = n-p;
N = min(m,n);

if (n~=size(L,2)) || (d<0), error('Incorrect dimensions for L.'), end

X = zeros(n,niter);
Alphas = zeros(niter,1);

normx0 = norm(x0);
X(:,1) = x0;
rGN = fun(X(:,1))-b;
rho = norm(rGN);

k = 0;
flag = 1;
while flag
	k = k+1;

	[fGN,jacGN] = fun(X(:,k));
	rGN = fGN-b;

	smalljac = 0;
	if curegsvd && (norm(jacGN,inf) < factgsvd)
		warning('small Jacobian!')
		smalljac = 1;
		[U,V,WI,C,S] = gsvd(jacGN/factgsvd,L);
	else
		[U,V,WI,C,S] = gsvd(jacGN,L);
	end
	WI = WI';	% WI=inv(W), with the formalism of the paper
	W = inv(WI);	% check if it's necessary
	if m>=n
		if n == 1
			c = C(1,1);
		else
			c = diag(C);
		end
		W1 = 0;
		W2 = W;
	else
		if m == 1
			c = C(1,n);
		else
			c = diag(C(:,n-m+1:n));
		end
		W1 = W(:,1:n-m);
		W2 = W(:,n-m+1:n);
	end
	c = c(1:r-d);		% excludes the ones
	%c = c(1:r);		% includes the ones - useless

	if p == 1
		s = S(1,1);
	else
		s = diag(S);
	end
	s = s(p-r+d+1:p);

	U = U(:,1:N);		% N = min(m,n)

	if smalljac
		beta = U'*(rGN/factgsvd);
	else
		beta = U'*rGN;
	end

	if mnflag
		z = WI*(X(:,k)-xbar);
		y = (c.*beta(1:r-d)+lam^2*s.^2.*z(n-r+1:p))./(c.^2+lam^2*s.^2);
	else
		y = (c.*beta(1:r-d))./(c.^2+lam^2*s.^2);
	end
	step = -W2*[y;beta(r-d+1:N)];

	% damped
	alpha = 2;
	flagd = 1;
	while flagd
		alpha = alpha/2;
		x1 = X(:,k) + alpha * step;
		rho1 = norm(fun(x1)-b);
		% Armijo-Goldstein principle
		if smalljac
			njs2 = (norm(jacGN*step)/factgsvd)^2;
		else
			njs2 = norm(jacGN*step)^2;
		end
		alphas = norm(alpha*step);
		flagd = (rho^2 - rho1^2) < (1/2)*alpha*njs2;
		flagalpha = alphas > pow2(minpowAG);
		flagd = flagd && flagalpha;
	end
	Alphas(k) = alpha;
	%     x1 = X(:,k) + alpha*step;
	if mnflag && (n>r)
		x1 = x1 - W1*z(1:n-r);
	end
	X(:,k+1) = x1;
	normx = norm(L*X(:,k+1));

	rho = norm(fun(x1)-b);

	diff = norm(X(:,k+1)-X(:,k)); % norm(x_k+1 - x_k)

	flagGN1 = (diff > 1e-8*max(1,normx)); % tol = 1e-8
	flagGN2 = (normx < 1e8*normx0);
	flagGN3 = (k < niter);
	%[flagGN1 flagGN2 flagGN3 flagalpha]
	flag = flagGN1 && flagGN2 && flagGN3 && flagalpha;
end

xGN = X(:,k+1);
X = X(:,1:k+1);
Alphas = Alphas(1:k);
