function [xGN, k, rho, X, Alphas] = tmngn( fun, b, x0, ell, niter, mnflag, xbar)
%TMNGN Gauss-Newton method regularized by TSVD
%   [x, k, rho, X, alpha] = tmngn( fun, b, x0, ell, niter, mnflag, xbar)
%   computes the minimum norm solution to the nonlinear least-square
%   problem min||fun(x)-b||.
%   where
%      fun    - function and Jacobian: [f,J]=fun(x)
%      b      - data vector
%      x0     - initial point
%      ell    - either rank or truncation parameter
%      niter  - max number of iterations
%      mnflag - 1 minimum norm solution, 0 standard solution
%      xbar   - a priori estimate of solution (only if mnflag=1)
%   and
%      x      - solution
%      k      - number of iterations performed
%      rho    - solution residual
%      X      - solution at each iteration
%      alpha  - step-lenghts for damped Gauss-Newton method

%   F. Pes and G. Rodriguez
%   University of Cagliari, Italy

% Last revised July 24, 2020

if nargin<7, xbar = zeros(size(x0)); end
if nargin<6, mnflag = 1; end
if nargin<5, niter = 40; end
if nargin<4, ell = inf; end

if mnflag==0 && ~(isequal(xbar,zeros(size(x0)))), error('xbar only if mnflag=1.'), end

minpowAG = -30;		% alpha > 2^minpowAG for Armijo-Goldstein

n = size(x0,1);
m = size(b,1);
if ~isfinite(ell)		% max value for ell
	ell = min(m,n);
end

X = zeros(n,niter);
Alphas = zeros(niter,1);

normx0 = norm(x0);
X(:,1) = x0;
rGN = fun(X(:,1))-b;
rho = norm(rGN);

k = 0;
flag = 1;
while flag
	k = k+1;

	[fGN,jacGN] = fun(X(:,k));
	rGN = fGN-b;

	[U,S,V] = svd(jacGN);
	if min(m,n)==1
		d = S(1,1);
	else
		d = diag(S);
	end

	if ell<n
		V2 = V(:,ell+1:n);
	else
		V2 = 0;
	end
	U = U(:,1:ell);
	V = V(:,1:ell);
	d = d(1:ell);

	%     step = - pinv(jacGN)*rGN;
	y = U'*rGN;
	y = -y./d;
	step = V*y;

	% damped
	alpha = 2;
	flagd = 1;
	while flagd
		alpha = alpha/2;
		x1 = X(:,k) + alpha * step;
		rho1 = norm(fun(x1)-b);
		% Armijo-Goldstein principle
		alphas = norm(alpha*step);
		flagd = (rho^2 - rho1^2) < (1/2)*alpha*norm(jacGN*step)^2;
		flagalpha = alphas > pow2(minpowAG);
		flagd = flagd && flagalpha;
	end
	Alphas(k) = alpha;
	%     x1 = X(:,k) + alpha*step; % already computed
	if mnflag
		x1 = x1 - V2*(V2'*(X(:,k)-xbar));
	end
	X(:,k+1) = x1;
	normx = norm(X(:,k+1));

	rho = norm(fun(x1)-b);

	diff = norm(X(:,k+1)-X(:,k)); % norm(x_k+1 - x_k)

	flagGN1 = (diff > 1e-8*max(1,normx)); % tol = 1e-8
	flagGN2 = (normx < 1e8*normx0);
	flagGN3 = (k < niter);
	%[flagGN1 flagGN2 flagGN3 flagalpha]
	flag = flagGN1 && flagGN2 && flagGN3 && flagalpha;
end

xGN = X(:,k+1);
X = X(:,1:k+1);
Alphas = Alphas(1:k);
